import re

import path
import pytest


@pytest.fixture()
def tmp_path(tmpdir):
    return path.Path(tmpdir)


@pytest.fixture
def test_path():
    this_dir = path.Path(__file__).parent
    return this_dir.abspath()



import attr
import collections


def increment(value):
    return str(int(value) + 1)


def reset(value):
    return "0"


class Part:
    def __init__(self, name, value, position):
        self.name = name
        self.value = value
        self.position = position

    def __repr__(self):
        return self.name + ":" + (self.value or "<none>")

    def __str__(self):
        return repr(self)


def parse_version(regex, version):
    match = regex.match(version)
    matchdict = match.groupdict()
    res = list()
    for name, value in matchdict.items():
        position = match.start(name)
        if position != -1:
            part = Part(name, value, position)
            res.append(part)
    res.sort(key=lambda x: x.position)
    return res


def get_parts(regex, version):
    match = regex.match(version)
    matchdict = match.groupdict()
    return matchdict.group()


class Bumper:
    def __init__(self, *, parse, serialize):
        self.parse = parse
        self.serialize = serialize

    def bump(self, version, part_name):
        known_parts = get_parts(version)
        parts = parse_version(self.parse, version)
        found = False
        for part in parts:
            if part.name == part_name:
                part.value = increment(part.value)
                found = True
            else:
                if found:
                    part.value = reset(part.value)
        for known_parts in known_part:
            pass

        template = self.serialize[0]
        for part in parts:
            template = template.replace("{%s}" % part.name, part.value)
        return template



import os
import re

import attr
import schema
import toml
import ui


@attr.s
class Config:
    current_version = attr.ib()
    tag_template = attr.ib(default=None)
    message_template = attr.ib(default=None)
    files = attr.ib(default=None)


@attr.s
class File:
    src = attr.ib()
    search = attr.ib(default=None)


class ValidTag():
    message = "tag_template should contain the string {new_version}"

    def validate(self, value):
        if "{new_version}" not in value:
            raise schema.SchemaError(self.message)
        return True


class ValidFile():
    message = "src should be an existing path"

    def validate(self, value):
        if not os.path.exists(value):
            raise schema.SchemaError(self.message)


def validate(config):
    file_schema = schema.Schema({
        "src": ValidFile(),
        schema.Optional("search"): str,
    })
    tbump_schema = schema.Schema(
      {
        "version":  {
          "current": str,
        },
        "git": {
          "message_template": str,
          "tag_template": ValidTag(),
        },
        "file": [file_schema],
        }
    )
    tbump_schema.validate(config)


def parse(cfg_path):
    parsed = None
    with cfg_path.open() as stream:
        parsed = toml.load(stream)

    validate(parsed)
    config = Config(
        current_version=parsed["version"]["current"]
    )
    config.tag_template = parsed["git"]["tag_template"]
    config.message_template = parsed["git"]["message_template"]

    config.files = list()
    for file_dict in parsed["file"]:
        file = File(
            src=file_dict["src"],
            search=file_dict.get("search"),
        )
        config.files.append(file)
    return config



import argparse
import os
import sys

import path
import schema
import ui

import tbump.config
from tbump.git import run_git

TBUMP_VERSION = "0.0.5"


def display_diffs(file_path, diffs):
    ui.info_2("Patching",
              ui.reset, ui.bold, file_path)
    for old, new in diffs:
        ui.info(ui.red, "-", old)
        ui.info(ui.green, "+", new)


def should_replace(line, old_string, search=None):
    if not search:
        return old_string in line
    else:
        return (old_string in line) and (search in line)


def replace_in_file(file_path, old_string, new_string, search=None):
    old_lines = file_path.lines(retain=False)
    diffs = list()
    new_lines = list()
    for old_line in old_lines:
        if should_replace(old_line, old_string, search):
            new_line = old_line.replace(old_string, new_string)
            diffs.append((old_line, new_line))
        else:
            new_line = old_line
        new_lines.append(new_line)
    if not diffs:
        ui.fatal("File", file_path, "did not change")
    display_diffs(file_path, diffs)
    file_path.write_lines(new_lines)


def check_dirty(working_path):
    rc, out = run_git(working_path, "status", "--porcelain", raises=False)
    if rc != 0:
        ui.fatal("git status failed")
    dirty = False
    for line in out.splitlines():
        # Ignore untracked files
        if not line.startswith("??"):
            dirty = True
    if dirty:
        ui.error("Repository is dirty")
        ui.info(out)
        sys.exit(1)


def commit(working_path, message):
    ui.info_2("Making bump commit")
    run_git(working_path, "add", ".")
    run_git(working_path, "commit", "--message", message)


def check_ref_does_not_exists(working_path, tag_name):
    rc, _ = run_git(working_path, "rev-parse", tag_name, raises=False)
    if rc == 0:
        ui.fatal("git ref", tag_name, "already exists")


def tag(working_path, tag):
    ui.info_2("Creating tag", tag)
    run_git(working_path, "tag", tag)


def get_current_branch(working_path):
    cmd = ("rev-parse", "--abbrev-ref", "HEAD")
    rc, out = run_git(working_path, *cmd, raises=False)
    if rc != 0:
        ui.fatal("Failed to get current ref")
    if out == "HEAD":
        ui.fatal("Not on any branch")
    return out


def get_tracking_ref(working_path):
    rc, out = run_git(working_path,
                      "rev-parse", "--abbrev-ref", "--symbolic-full-name", "@{u}",
                      raises=False)
    if rc != 0:
        ui.fatal("Failed to get tracking ref")
    return out


def parse_config():
    config = tbump.config.parse(path.Path("tbump.toml"))
    return config


def bump_version(config, new_version):
    current_version = config.current_version
    for file in config.files:
        file_path = path.Path(file.src)
        to_search = None
        if file.search:
            to_search = file.search.format(current_version=current_version)
        replace_in_file(file_path, current_version, new_version, search=to_search)
    replace_in_file(path.Path("tbump.toml"), current_version, new_version)


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("new_version")
    parser.add_argument("-C", "--cwd", dest="working_dir")
    parser.add_argument("--non-interactive", dest="interactive", action="store_false")
    parser.add_argument("--version", action="version", version=TBUMP_VERSION)
    args = parser.parse_args(args=args)
    working_dir = args.working_dir
    new_version = args.new_version
    if working_dir:
        os.chdir(working_dir)
    try:
        config = parse_config()
    except IOError as io_error:
        ui.fatal("Could not read config file:", io_error)
    except Exception as e:
        ui.fatal("Invalid config:",  e)
    ui.info_1(
            "Bumping from",
            ui.reset, ui.bold, config.current_version,
            ui.reset, "to",
            ui.reset, ui.bold, new_version)
    working_path = path.Path.getcwd()

    check_dirty(working_path)
    branch_name = get_current_branch(working_path)
    tracking_ref = get_tracking_ref(working_path)
    remote_name, remote_branch = tracking_ref.split("/", maxsplit=1)

    tag_name = config.tag_template.format(new_version=new_version)
    check_ref_does_not_exists(working_path, tag_name)

    bump_version(config, new_version)

    message = config.message_template.format(new_version=new_version)
    commit(working_path, message)

    tag(working_path, tag_name)

    if args.interactive:
        answer = ui.ask_yes_no("OK to push", default=False)
        if answer:
            run_git(working_path, "push", remote_name, remote_branch, verbose=True)
            run_git(working_path, "push", remote_name, tag_name, verbose=True)



import re

import path
import pytest
import schema

import tbump.config


def test_happy_parse(monkeypatch):
    this_dir = path.Path(__file__).parent
    monkeypatch.chdir(this_dir)
    config = tbump.config.parse(path.Path("tbump.toml"))
    foo_json = tbump.config.File(
        src="package.json",
        search='"version": "{current_version}"'
    )
    version_txt = tbump.config.File(src="VERSION")

    assert config.files == [
        foo_json,
        version_txt,
    ]

    assert config.current_version == "1.2.41"


def test_wrong_syntax():
    this_dir = path.Path(__file__).parent
    invalid_path = this_dir.abspath().joinpath("invalid.toml")

    with pytest.raises(schema.SchemaError) as e:
        tbump.config.parse(invalid_path)
    assert "should contain" in e.value.args[0]



import subprocess

import ui


def run_git(working_path, *cmd, raises=True, verbose=False):
    git_cmd = list(cmd)
    git_cmd.insert(0, "git")
    options = dict()
    if not raises:
        options["stdout"] = subprocess.PIPE
        options["stderr"] = subprocess.STDOUT

    if verbose:
        ui.info_3(*git_cmd)
    process = subprocess.Popen(git_cmd, cwd=working_path, **options)

    if raises:
        process.wait()
    else:
        out, _ = process.communicate()
        out = out.decode("utf-8")

    returncode = process.returncode
    if raises:
        if returncode != 0:
            ui.fatal(" ".join(git_cmd), "failed")
    else:
        if out.endswith('\n'):
            out = out.strip('\n')
        ui.debug(ui.lightgray, "[%i]" % returncode, ui.reset, out)
        return returncode, out



import sys
from setuptools import setup, find_packages

if sys.version_info.major < 3:
    sys.exit("Error: Please upgrade to Python3")


def get_long_description():
    with open("README.rst") as fp:
        return fp.read()


setup(name="tbump",
      version="0.0.5",
      description="Bump software releases",
      long_description=get_long_description(),
      url="https://github.com/SuperTanker/tbump",
      author="Dimitri Merejkowsky",
      packages=find_packages(),
      include_package_data=True,
      install_requires=[
        "attrs",
        "path.py",
        "python-cli-ui",
        "schema",
        "toml",
      ],
      classifiers=[
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
      ],
      entry_points={
        "console_scripts": [
          "tbump = tbump.main:main",
         ]
      }
      )



import subprocess
import textwrap

import toml
import path
import pytest
from ui.tests.conftest import message_recorder

import tbump.main
import tbump.git


def assert_in_file(file_name, expected_line):
    file_path = path.Path(file_name)
    for line in file_path.lines():
        if expected_line in line:
            return
    assert False, "No line found matching %s" % expected_line


def setup_test(test_path, tmp_path, monkeypatch):
    src_path = tmp_path.joinpath("src")
    src_path.mkdir()
    test_path.joinpath("tbump.toml").copy(src_path)
    src_path.joinpath("VERSION").write_text("1.2.41")
    src_path.joinpath("package.json").write_text(textwrap.dedent("""
    {
       "name": "foo",
       "version": "1.2.41",
       "dependencies": {
         "some-dep": "1.3",
         "other-dep": "1.2.41"
       }
    }
    """))
    tbump.git.run_git(src_path, "init")
    tbump.git.run_git(src_path, "add", ".")
    tbump.git.run_git(src_path, "commit", "--message", "initial commit")
    setup_remote(tmp_path)
    return src_path


def setup_remote(tmp_path):
    git_path = tmp_path.joinpath("git")
    git_path.mkdir()
    remote_path = git_path.joinpath("repo.git")
    remote_path.mkdir()
    tbump.git.run_git(remote_path, "init", "--bare")

    src_path = tmp_path.joinpath("src")
    tbump.git.run_git(src_path, "remote", "add", "origin", remote_path)
    tbump.git.run_git(src_path, "push", "-u", "origin", "master")


def test_replaces(tmp_path, test_path, monkeypatch):
    src_path = setup_test(test_path, tmp_path, monkeypatch)
    tbump.main.main(["-C", src_path, "1.2.42-alpha-1", "--non-interactive"])

    toml_path = src_path.joinpath("tbump.toml")
    new_toml = toml.loads(toml_path.text())
    assert new_toml["version"]["current"] == "1.2.42-alpha-1"

    assert_in_file("package.json", '"version": "1.2.42-alpha-1"')
    assert_in_file("package.json", '"other-dep": "1.2.41"')


def test_commit_and_tag(tmp_path, test_path, monkeypatch):
    src_path = setup_test(test_path, tmp_path, monkeypatch)
    tbump.main.main(["-C", src_path, "1.2.42-alpha-1", "--non-interactive"])

    rc, out = tbump.git.run_git(src_path, "log", "--oneline", raises=False)
    assert rc == 0
    assert "Bump to 1.2.42-alpha-1" in out

    rc, out = tbump.git.run_git(src_path, "tag", "--list", raises=False)
    assert rc == 0
    assert out == "v1.2.42-alpha-1"


def test_abort_if_dirty(tmp_path, test_path, monkeypatch, message_recorder):
    src_path = setup_test(test_path, tmp_path, monkeypatch)
    src_path.joinpath("VERSION").write_text("unstaged changes\n", append=True)

    with pytest.raises(SystemExit) as e:
        tbump.main.main(["-C", src_path, "1.2.42-alpha-1", "--non-interactive"])
    assert message_recorder.find("dirty")


def test_abort_if_tag_exists(tmp_path, test_path, monkeypatch, message_recorder):
    src_path = setup_test(test_path, tmp_path, monkeypatch)
    tbump.git.run_git(src_path, "tag", "v1.2.42")

    with pytest.raises(SystemExit) as e:
        tbump.main.main(["-C", src_path, "1.2.42", "--non-interactive"])
    assert message_recorder.find("1.2.42 already exists")


def test_abort_if_file_does_not_change(tmp_path, test_path, monkeypatch, message_recorder):
    src_path = setup_test(test_path, tmp_path, monkeypatch)
    invalid_src = src_path.joinpath("foo.txt")
    invalid_src.write_text("this is foo")
    tbump_path = src_path.joinpath("tbump.toml")
    tbump_path.write_text("""\
    [[file]]
    src = "foo.txt"
    """, append=True)
    tbump.git.run_git(src_path, "add", ".")
    tbump.git.run_git(src_path, "commit", "--message", "add foo.txt")

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", src_path, "1.2.42", "--non-interactive"])
    assert "foo.txt did not change"


def test_interactive_push(tmp_path, test_path, monkeypatch, message_recorder, mock):
    src_path = setup_test(test_path, tmp_path, monkeypatch)
    ask_mock = mock.patch("ui.ask_yes_no")
    ask_mock.return_value = True
    tbump.main.main(["-C", src_path, "1.2.42"])
    ask_mock.assert_called_with("OK to push", default=False)
    rc, out = tbump.git.run_git(src_path, "ls-remote", raises=False)
    assert rc == 0
    assert "tags/v1.2.42" in out
